%
% @(#)inout.pl
% Input/Output built-in predicates for Prolog Cafe
% 

% @author Mutsunori Banbara 
%         banbara@pascal.seg.kobe-u.ac.jp
%         Nara National College of Technology
% @author Naoyuki Tamura 
%         tamura@kobe-u.ac.jp
%         Faculty of Engineering, Kobe University

% JJC: change java_object to cs_object

:- dynamic '$current_op0'/3.
:- resource '$tokens'/1.       % Prolog Cafe specific

/*===============================================================
	Input of Terms
===============================================================*/
readint(N) :-
	'$skip_spaces'(C),
	'$readint'(C, N).

'$readint'(C, N) :-
	C =:= "-", !,
	get0(C1),
	'$readint1'(C1, 0, N1),
	N is -N1.
'$readint'(C, N) :-
	'$readint1'(C, 0, N).

'$readint1'(C, N0, N) :-
	C >= "0", C =< "9",
	!,
	N1 is 10*N0+C-"0",
	get0(C1),
	'$readint1'(C1, N1, N).
'$readint1'(10, N, N) :- !.			% end of line
'$readint1'(_, N, N) :- !,
	skip(10).				% skip to end of line

'$skip_spaces'(C) :- get0(C0), '$skip_spaces1'(C0, C).

'$skip_spaces1'( 9, C) :- !, '$skip_spaces'(C).
'$skip_spaces1'(10, C) :- !, '$skip_spaces'(C).
'$skip_spaces1'(12, C) :- !, '$skip_spaces'(C).
'$skip_spaces1'(13, C) :- !, '$skip_spaces'(C).
'$skip_spaces1'(32, C) :- !, '$skip_spaces'(C).
'$skip_spaces1'(C, C).

readline(Str) :- get0(C), '$readline'(C, Str).

'$readline'(-1, [-1]) :- !.			% end_of_file
'$readline'(10, []) :- !.			% LF means end_of_line
'$readline'(13, Str) :- !, readline(Str).	% CR is ignored
'$readline'(C, [C|Str]) :- !, readline(Str).

%
%   read_token(Token) reads one token from the input, and returns
%     error('message'), end_of_file, 
%     '.', ' ', '(', ')', '[', ']', '{', '}', ',', '|', 
%     number(value), atom('name'), var('name'), string("chars")
%
read_token(Token) :-
	'$read_token0'(Type, Token0),
	'$read_token1'([Type], Token0, Token).

'$read_token1'([-2], T, error(T)) :- !.		% error('message')
'$read_token1'("I", T, number(T)) :- !.		% number(value)
'$read_token1'("A", T, atom(T)) :- !.		% atom('name')
'$read_token1'("V", T, var(T)) :- !.		% var('name')
'$read_token1'("S", T, string(S)) :- !, name(T, S).	% string("chars")
'$read_token1'(Type, T, T) :- !.		% others

%
%   read_tokens(Tokens, Vs) reads tokens from the input
%     until full-stop-mark ('.') or end_of_file,
%     and returns a list of tokens.
%     Token for a variable has a form of var(Name,Variable).
%     Vs is a list of Name=Variable pairs.
%
read_tokens(Tokens, Vs) :-
	'$read_tokens'(Tokens, Vs, []),
	!.

'$read_tokens'(Tokens, Vs, VI) :-
	read_token(Token),
	'$read_tokens1'(Token, Tokens, Vs, VI).

'$read_tokens1'(error(Message), [], Vs, _) :- !,
	write('SYNTAX ERROR'), nl, write('** '),
	write(Message), write(' **'), nl,
	'$read_tokens_until_fullstop',
	fail.
'$read_tokens1'(end_of_file, [end_of_file,'.'], [], _) :- !.
'$read_tokens1'('.', ['.'], [], _) :- !.


'$read_tokens1'(var('_'), [var('_',V)|Tokens], ['_'=V|Vs], VI0) :- !,
	'$read_tokens'(Tokens, Vs, ['_'=V|VI0]).

'$read_tokens1'(var(Name), [var(Name,V)|Tokens], Vs, VI) :-
	'$mem_pair'(Name=V, VI), !,
	'$read_tokens'(Tokens, Vs, VI).

'$read_tokens1'(var(Name), [var(Name,V)|Tokens], [Name=V|Vs], VI0) :- !,
	'$read_tokens'(Tokens, Vs, [Name=V|VI0]).

'$mem_pair'(X1=V1, [X2=V2|_]) :- X1 == X2, !, V1 = V2.
% $B$h$/$o$+$i$s!*(B
'$mem_pair'(X, [_|L]) :- '$member'(X, L).
%'$mem_pair'(X, [_|L]) :- '$mem_pair'(X, L).

'$read_tokens1'(Token, [Token|Tokens], Vs, VI) :-
	'$read_tokens'(Tokens, Vs, VI).

'$read_tokens_until_fullstop' :-
	read_token(Token),
	'$read_tokens_until_fullstop'(Token).

'$read_tokens_until_fullstop'(end_of_file) :- !.
'$read_tokens_until_fullstop'('.') :- !.
'$read_tokens_until_fullstop'(_) :-
	read_token(Token),
	'$read_tokens_until_fullstop'(Token).

%
%   read(X) reads a term from the input
%
read(X) :-
	read_tokens(Tokens, _),
	parse_tokens(X, Tokens),
	!.

read(Stream, X) :-
	current_input(In),
	set_input(Stream),
	read(X),
	set_input(In).

read_with_variables(X, Vs) :-
	read_tokens(Tokens, Vs),
	parse_tokens(X, Tokens),
	!.

%parse_tokens(X, Tokens) :-
%	'$parse_tokens'(X, 1201, Tokens, ['.']),
%	!.

parse_tokens(X, Tokens) :-
	('$tokens'(Tokens) => '$parse_tokens'(X, 1201, Tokens, ['.'])),
	!.

%
%  '$parse_tokens'(X, Prec0) parses the input whose precedecence =< Prec.
%
'$parse_tokens'(X, Prec0) -->
	'$parse_tokens_skip_spaces',
	'$parse_tokens1'(Prec0, X1, Prec1),
	!,
	'$parse_tokens_skip_spaces',
	'$parse_tokens2'(Prec0, X1, Prec1, X, Prec),
	!.

'$parse_tokens1'(Prec0, X1, Prec1) -->
	'$parse_tokens_peep_next'(Next),
	{'$parse_tokens_is_starter'(Next)},
	!,
	'$parse_tokens_before_op'(Prec0, X1, Prec1).
'$parse_tokens1'(_, _, _) -->
	'$parse_tokens_peep_next'(Next),
	'$parse_tokens_error'([Next,cannot,start,an,expression]).

'$parse_tokens2'(Prec0, X, Prec, X, Prec) -->
	'$parse_tokens_peep_next'(Next),
	{'$parse_tokens_is_terminator'(Next)},
	{Prec =< Prec0},
	!.
'$parse_tokens2'(Prec0, X1, Prec1, X, Prec) -->
	'$parse_tokens_peep_next'(Next),
	{'$parse_tokens_is_post_in_op'(Next)},
	!,
	'$parse_tokens_post_in_ops'(Prec0, X1, Prec1, X, Prec).
'$parse_tokens2'(_, _, _, _, _) --> 
	'$parse_tokens_error'([operator,expected,after,expression]).

%
%  '$parse_tokens_before_op'(Prec0, X, Prec)
%  parses the input until infix or postfix operator,
%  and returns X and Prec
%
'$parse_tokens_before_op'(Prec0, X, Prec) --> [' '], !,
	'$parse_tokens_before_op'(Prec0, X, Prec).
'$parse_tokens_before_op'(_, end_of_file, 0) --> [end_of_file], !.
'$parse_tokens_before_op'(_, N, 0) --> [number(N)], !.
'$parse_tokens_before_op'(_, N, 0) -->
	[atom('-')], [number(N0)], !, {N is -N0}.
'$parse_tokens_before_op'(_, V, 0) --> [var(_,V)], !.
'$parse_tokens_before_op'(_, S, 0) --> [string(S)], !.
'$parse_tokens_before_op'(_, X, 0) --> ['('], !,
	'$parse_tokens'(X, 1201),
	'$parse_tokens_expect'(')').
'$parse_tokens_before_op'(_, X, 0) --> ['{'], !,
	'$parse_tokens_skip_spaces',
	'$parse_tokens_brace'(X).
'$parse_tokens_before_op'(_, X, 0) --> ['['], !,
	'$parse_tokens_skip_spaces',
	'$parse_tokens_list'(X).
'$parse_tokens_before_op'(_, X, 0) -->
	[atom(F)], ['('],
	!,
	'$parse_tokens_skip_spaces',
	'$parse_tokens_args'(Args),
	{X =.. [F|Args]}.
'$parse_tokens_before_op'(Prec0, X, PrecOp) -->
	[atom(F)], {current_op(PrecOp,fx,F)}, {PrecOp =< Prec0},
	'$parse_tokens_skip_spaces',
	'$parse_tokens_peep_next'(Next),
	{'$parse_tokens_is_starter'(Next)},
	{\+ '$parse_tokens_is_post_in_op'(Next)},
	!,
        {Prec1 is PrecOp - 1},
        '$parse_tokens'(Arg, Prec1),
	{functor(X, F, 1)},
	{arg(1, X, Arg)}.
'$parse_tokens_before_op'(Prec0, X, PrecOp) -->
	[atom(F)], {current_op(PrecOp,fy,F)}, {PrecOp =< Prec0},
	'$parse_tokens_skip_spaces',
	'$parse_tokens_peep_next'(Next),
	{'$parse_tokens_is_starter'(Next)},
	{\+ '$parse_tokens_is_post_in_op'(Next)},
	!,
        '$parse_tokens'(Arg, PrecOp),
	{functor(X, F, 1)},
	{arg(1, X, Arg)}.
'$parse_tokens_before_op'(_, A, 0) -->
	[atom(A)].

'$parse_tokens_brace'('{}') --> ['}'], !.
'$parse_tokens_brace'(X) -->
	'$parse_tokens'(X1, 1201),
	'$parse_tokens_expect'('}'),
	{X = {X1}}.

'$parse_tokens_list'('[]') --> [']'], !.
'$parse_tokens_list'([X|Xs]) -->
	'$parse_tokens'(X, 999),
	'$parse_tokens_skip_spaces',
	'$parse_tokens_list_rest'(Xs).

'$parse_tokens_list_rest'(Xs) --> ['|'], !,
	'$parse_tokens'(Xs, 999),
	'$parse_tokens_expect'(']').
'$parse_tokens_list_rest'([X|Xs]) --> [','], !,
	'$parse_tokens'(X, 999),
	'$parse_tokens_skip_spaces',
	'$parse_tokens_list_rest'(Xs).
'$parse_tokens_list_rest'('[]') -->
	'$parse_tokens_expect'(']').

'$parse_tokens_args'('[]') --> [')'], !.
'$parse_tokens_args'([X|Xs]) -->
	'$parse_tokens'(X, 999),
	'$parse_tokens_skip_spaces',
	'$parse_tokens_args_rest'(Xs).

'$parse_tokens_args_rest'([X|Xs]) --> [','], !,
	'$parse_tokens'(X, 999),
	'$parse_tokens_skip_spaces',
	'$parse_tokens_args_rest'(Xs).
'$parse_tokens_args_rest'('[]') -->
	'$parse_tokens_expect'(')').

%
%  '$parse_tokens_post_in_op'(Prec0, X1, Prec1, X, Prec)
%  parses the input beginning from infix or postfix operator,
%  and returns X and Prec
%
'$parse_tokens_post_in_ops'(Prec0, X1, Prec1, X, Prec) -->
	'$parse_tokens_skip_spaces',
	[Op],
	'$parse_tokens_op'(Op, Prec0, X1, Prec1, X2, Prec2),
	'$parse_tokens_post_in_ops'(Prec0, X2, Prec2, X, Prec).
'$parse_tokens_post_in_ops'(Prec0, X, Prec, X, Prec) --> 
	{Prec =< Prec0}.

'$parse_tokens_op'(',', Prec0, X1, Prec1, X, PrecOp) --> !,
	'$parse_tokens_op'(atom(','), Prec0, X1, Prec1, X, PrecOp).
'$parse_tokens_op'('|', Prec0, X1, Prec1, X, PrecOp) --> !,
	'$parse_tokens_op'(atom(';'), Prec0, X1, Prec1, X, PrecOp).
'$parse_tokens_op'(atom(Op), Prec0, X1, Prec1, X, PrecOp) -->
	{current_op(PrecOp, xf, Op)}, {PrecOp =< Prec0},
	{Prec1 < PrecOp},
	{functor(X, Op, 1)},
	{arg(1, X, X1)}.
'$parse_tokens_op'(atom(Op), Prec0, X1, Prec1, X, PrecOp) -->
	{current_op(PrecOp, yf, Op)}, {PrecOp =< Prec0},
	{Prec1 =< PrecOp},
	{functor(X, Op, 1)},
	{arg(1, X, X1)}.
'$parse_tokens_op'(atom(Op), Prec0, X1, Prec1, X, PrecOp) -->
	{current_op(PrecOp, xfx, Op)}, {PrecOp =< Prec0},
	{Prec1 < PrecOp},
	{Prec2 is PrecOp - 1},
	'$parse_tokens'(X2, Prec2),
	!,
	{functor(X, Op, 2)},
	{arg(1, X, X1)},
	{arg(2, X, X2)}.
'$parse_tokens_op'(atom(Op), Prec0, X1, Prec1, X, PrecOp) -->
	{current_op(PrecOp, xfy, Op)}, {PrecOp =< Prec0},
	{Prec1 < PrecOp},
	{Prec2 is PrecOp},
	'$parse_tokens'(X2, Prec2),
	!,
	{functor(X, Op, 2)},
	{arg(1, X, X1)},
	{arg(2, X, X2)}.
'$parse_tokens_op'(atom(Op), Prec0, X1, Prec1, X, PrecOp) -->
	{current_op(PrecOp, yfx, Op)}, {PrecOp =< Prec0},
	{Prec1 =< PrecOp},
	{Prec2 is PrecOp - 1},
	'$parse_tokens'(X2, Prec2),
	!,
	{functor(X, Op, 2)},
	{arg(1, X, X1)},
	{arg(2, X, X2)}.

'$parse_tokens_is_starter'(end_of_file).
'$parse_tokens_is_starter'('(').
'$parse_tokens_is_starter'('[').
'$parse_tokens_is_starter'('{').
'$parse_tokens_is_starter'(number(_)).
'$parse_tokens_is_starter'(atom(_)).
'$parse_tokens_is_starter'(var(_,_)).
'$parse_tokens_is_starter'(string(_)).

'$parse_tokens_is_terminator'(')').
'$parse_tokens_is_terminator'(']').
'$parse_tokens_is_terminator'('}').
'$parse_tokens_is_terminator'('.').

'$parse_tokens_is_post_in_op'(',') :- !.
'$parse_tokens_is_post_in_op'('|') :- !.
'$parse_tokens_is_post_in_op'(atom(Op)) :-
	current_op(_, Type, Op),
	'$parse_tokens_post_in_type'(Type),
	!.

'$parse_tokens_post_in_type'(xfx).
'$parse_tokens_post_in_type'(xfy).
'$parse_tokens_post_in_type'(yfx).
'$parse_tokens_post_in_type'(xf).
'$parse_tokens_post_in_type'(yf).

'$parse_tokens_expect'(Token) -->
	'$parse_tokens_skip_spaces',
	[Token],
	!.
'$parse_tokens_expect'(Token) -->
	'$parse_tokens_error'([Token,expected]).

'$parse_tokens_skip_spaces' --> [' '], !, '$parse_tokens_skip_spaces'.
'$parse_tokens_skip_spaces' --> [].

'$parse_tokens_peep_next'(Next, S, S) :- S = [Next|_].

'$parse_tokens_error'(Message, S0, S) :-
	write('SYNTAX ERROR'), nl, write('** '),
	'$parse_tokens_write_message'(Message), write(' **'), nl,
	'$parse_tokens_error1'([], S0),
	'$tokens'(Tokens),
	'$parse_tokens_error1'(Tokens, S0),
	fail.

'$parse_tokens_error1'([], _) :- !.
'$parse_tokens_error1'(Tokens, S0) :- Tokens == S0, !,
	nl, write('** here **'), nl,
	'$parse_tokens_error1'(Tokens, []), nl.
'$parse_tokens_error1'([Token|Tokens], S0) :-
	'$parse_tokens_error2'(Token),
	'$parse_tokens_error1'(Tokens, S0).

'$parse_tokens_error2'(number(X)) :- !, write(X).
'$parse_tokens_error2'(atom(X)) :- !, writeq(X).
'$parse_tokens_error2'(var(X,_)) :- !, write(X).
'$parse_tokens_error2'(string(X)) :- !,
	write('"'), '$parse_tokens_write_string'(X), write('"').
'$parse_tokens_error2'(X) :- write(X).

'$parse_tokens_write_string'([]).
'$parse_tokens_write_string'([C|Cs]) :- [C] = """", !,
	put(C), put(C), '$parse_tokens_write_string'(Cs).
'$parse_tokens_write_string'([C|Cs]) :- 
	put(C), '$parse_tokens_write_string'(Cs).

'$parse_tokens_write_message'([]).
'$parse_tokens_write_message'([X|Xs]) :-
	write(X), write(' '), '$parse_tokens_write_message'(Xs).

/*===============================================================
	Output of Terms
===============================================================*/
display(X) :-
	current_output(Out),
	set_output(user),
	'$display'(X),
	set_output(Out).

'$display'(X) :- atomic(X), !, '$fast_write'(X).
'$display'(X) :- cs_object(X), !, '$fast_write'(X).
%'$display'(X) :- '$write_term'(X, [ignore_ops]), !.
'$display'(X) :- '$write_term'(X, [ignore_ops]), fail.
'$display'(_).

write(X) :- atomic(X), !, '$fast_write'(X).
write(X) :- cs_object(X), !, '$fast_write'(X).
write(X) :- '$write_term'(X, [numbervars]), !.
%write(X) :- '$write_term'(X, [numbervars]), fail.
write(_).

write(Stream, X) :-
	current_output(Out),
	set_output(Stream),
	write(X),
	set_output(Out).

writeq(X) :- atomic(X), !, '$fast_writeq'(X).
writeq(X) :- cs_object(X), !, '$fast_writeq'(X).
%writeq(X) :- '$write_term'(X, [quoted, numbervars]), !.
writeq(X) :- '$write_term'(X, [quoted, numbervars]), fail.
writeq(_).

writeq(Stream, X) :-
	current_output(Out),
	set_output(Stream),
	writeq(X),
	set_output(Out).

write_canonical(X) :- atomic(X), !, '$fast_writeq'(X).
write_canonical(X) :- cs_object(X), !, '$fast_writeq'(X).
%write_canonical(X) :- '$write_term'(X, [quoted, ignore_ops]), !.
write_canonical(X) :- '$write_term'(X, [quoted, ignore_ops]), fail.
write_canonical(_).

write_canonical(Stream, X) :-
	current_output(Out),
	set_output(Stream),
	write_canonical(X),
	set_output(Out).

'$atom_type'(X, alpha ) :- '$atom_type0'(X, 0), !.
'$atom_type'(X, symbol) :- '$atom_type0'(X, 1), !.
'$atom_type'(X, punct ) :- '$atom_type0'(X, 2), !.
'$atom_type'(X, other ) :- '$atom_type0'(X, 3), !.

'$write_space_if_needed'(punct, _) :- !.
'$write_space_if_needed'(X, X) :- !, put(" ").
'$write_space_if_needed'(other, alpha) :- !, put(" ").
'$write_space_if_needed'(_, _).

% '$write_term'(Term, Style)
%  Style = [E|Es]
%  E = "quoted" or "numbervars" or "ignore_ops"

'$write_term'(Term, Style) :-
	'$write_term0'(Term, 1200, punct, _, Style),
	!.

'$write_term0'(Term, Prec, Type0, alpha, _) :- var(Term), !,
	'$write_space_if_needed'(Type0, alpha),
	'$fast_write'(Term).

%%% add 29/March/1999 by banbara
'$write_term0'(Term, Prec, Type0, alpha, _) :- cs_object(Term), !,
	'$write_space_if_needed'(Type0, alpha),
	'$fast_write'(Term).

'$write_term0'(Term, Prec, Type0, alpha, Style) :-
	Term = '$VAR'(VN), integer(VN), VN >= 0,
	'$member'(numbervars, Style),
	!,
	'$write_space_if_needed'(Type0, alpha),
	'$write_VAR'(VN).
%%%-->
'$write_VAR'(VN) :- VN < 26, !,
	Letter is VN mod 26 + "A",
	put(Letter).
'$write_VAR'(VN) :-
	Letter is VN mod 26 + "A",
	put(Letter),
	Rest is VN//26,
	'$fast_write'(Rest).
%%%<--

'$write_term0'(Term, Prec, Type0, alpha, _) :- integer(Term), Term < 0, !,
	'$write_space_if_needed'(Type0, symbol),
	'$fast_write'(Term).

'$write_term0'(Term, Prec, Type0, alpha, _) :- integer(Term), !,
	'$write_space_if_needed'(Type0, alpha),
	'$fast_write'(Term).

%'$write_term0'(Term, Prec, Type0, punct) :-
%	atom(Term), current_op(PrecOp, OpType, Term),
%	(OpType = fx; OpType = fy),
%	PrecOp =< Prec,
%	!,
%	'$write_space_if_needed'(Type0, punct),
%	put("("),
%	'$write_atom'(Term, punct, _),
%	put(")").

'$write_term0'(Term, Prec, Type0, Type, Style) :- atom(Term), !,
	'$write_atom'(Term, Type0, Type, Style).
%%%-->
'$write_atom'(Atom, Type0, Type, Style) :- '$member'(quoted, Style), !,
	'$atom_type'(Atom, Type), 
	'$write_space_if_needed'(Type0, Type),
	'$fast_writeq'(Atom).

'$write_atom'(Atom, Type0, Type, _) :-
	'$atom_type'(Atom, Type), 
	'$write_space_if_needed'(Type0, Type),
	'$fast_write'(Atom).
%%%<--

'$write_term0'(Term, Prec, Type0, Type, Style) :-
	\+ '$member'(ignore_ops, Style),
	'$write_is_operator'(Term, Op, Args, OpType),
	!,
	'$write_term_op'(Op, OpType, Args, Prec, Type0, Type, Style).
%%%-->
'$write_is_operator'(Term, Op, Args, OpType) :-
	functor(Term, Op, Arity),
	'$write_op_type'(Arity, OpType),
	current_op(_, OpType, Op),
	Term =.. [_|Args],
	!.

'$write_op_type'(1, fx).
'$write_op_type'(1, fy).
'$write_op_type'(1, xf).
'$write_op_type'(1, yf).
'$write_op_type'(2, xfx).
'$write_op_type'(2, xfy).
'$write_op_type'(2, yfx).
%%%<--

'$write_term0'(Term, Prec, Type0, punct, Style) :-
	Term = [_|_],
	\+ '$member'(ignore_ops, Style),
	!,
	'$write_space_if_needed'(Type0, punct),
	put("["),
	'$write_term_list_args'(Term, punct, _, Style),
	put("]").
%%%-->
'$write_term_list_args'([A|As], Type0, Type, Style) :- nonvar(As), As = [_|_], !,
	'$write_term0'(A, 999, Type0, Type1, Style),
	'$write_space_if_needed'(Type1, punct),
	put(","),
	'$write_term_list_args'(As, punct, Type, Style).
'$write_term_list_args'([A|As], Type0, Type, Style) :- nonvar(As), As = [], !,
	'$write_term0'(A, 999, Type0, Type, Style).
'$write_term_list_args'([A|As], Type0, Type, Style) :-
	'$write_term0'(A, 999, Type0, Type1, Style),
	'$write_space_if_needed'(Type1, punct),
	put("|"),
	'$write_term0'(As, 999, punct, Type, Style).
%%%<--

'$write_term0'(Term, Prec, Type0, Type, Style) :-
	Term = {Term1},
	\+ '$member'(ignore_ops, Style),
	!,
	'$write_space_if_needed'(Type0, punct),
	put("{"),
	'$write_term0'(Term1, 1200, punct, _, Style),
	put("}").

'$write_term0'(Term, Prec, Type0, punct, Style) :-
	Term =.. [F|Args],
	'$write_atom'(F, Type0, _, Style),
	put("("),
	'$write_term_args'(Args, punct, _, Style),
	put(")").
%%%-->
'$write_term_args'([], Type, Type, _) :- !.
'$write_term_args'([A], Type0, Type, Style) :- !,
	'$write_term0'(A, 999, Type0, Type, Style).
'$write_term_args'([A|As], Type0, Type, Style) :- !,
	'$write_term0'(A, 999, Type0, Type1, Style),
	'$write_space_if_needed'(Type1, punct),
	put(","),
	'$write_term_args'(As, punct, Type, Style).
%%%<--

'$write_term_op'(Op, OpType, Args, Prec, Type0, punct, Style) :-
	current_op(PrecOp, OpType, Op),
	PrecOp > Prec,
	!,
	'$write_space_if_needed'(Type0, punct),
	put("("),
	'$write_term_op1'(Op, OpType, Args, PrecOp, punct, _, Style),
	put(")").
'$write_term_op'(Op, OpType, Args, Prec, Type0, Type, Style) :-
	current_op(PrecOp, OpType, Op),
	'$write_term_op1'(Op, OpType, Args, PrecOp, Type0, Type, Style).

'$write_term_op1'(Op, fx, [A1], PrecOp, Type0, Type, Style) :- !,
	'$write_atom'(Op, Type0, Type1, Style),
	Prec1 is PrecOp - 1,
	'$write_term0'(A1, Prec1, Type1, Type, Style).
'$write_term_op1'(Op, fy, [A1], PrecOp, Type0, Type, Style) :- !,
	'$write_atom'(Op, Type0, Type1, Style),
	Prec1 is PrecOp,
	'$write_term0'(A1, Prec1, Type1, Type, Style).
'$write_term_op1'(Op, xf, [A1], PrecOp, Type0, Type, Style) :- !,
	Prec1 is PrecOp - 1,
	'$write_term0'(A1, Prec1, Type0, Type1, Style),
	'$write_atom'(Op, Type1, Type, Style).
'$write_term_op1'(Op, yf, [A1], PrecOp, Type0, Type, Style) :- !,
	Prec1 is PrecOp,
	'$write_term0'(A1, Prec1, Type0, Type1, Style),
	'$write_atom'(Op, Type1, Type, Style).
'$write_term_op1'(Op, xfx, [A1,A2], PrecOp, Type0, Type, Style) :- !,
	Prec1 is PrecOp - 1,
	Prec2 is PrecOp - 1,
	'$write_term0'(A1, Prec1, Type0, Type1, Style),
	'$write_term_infix_op'(Op, Type1, Type2, Style),
	'$write_term0'(A2, Prec2, Type2, Type, Style).
'$write_term_op1'(Op, xfy, [A1,A2], PrecOp, Type0, Type, Style) :- !,
	Prec1 is PrecOp - 1,
	Prec2 is PrecOp,
	'$write_term0'(A1, Prec1, Type0, Type1, Style),
	'$write_term_infix_op'(Op, Type1, Type2, Style),
	'$write_term0'(A2, Prec2, Type2, Type, Style).
'$write_term_op1'(Op, yfx, [A1,A2], PrecOp, Type0, Type, Style) :- !,
	Prec1 is PrecOp,
	Prec2 is PrecOp - 1,
	'$write_term0'(A1, Prec1, Type0, Type1, Style),
	'$write_term_infix_op'(Op, Type1, Type2, Style),
	'$write_term0'(A2, Prec2, Type2, Type, Style).
	
'$write_term_infix_op'(',', Type0, punct, _) :- !,
	'$write_space_if_needed'(Type0, punct),
	put(",").
'$write_term_infix_op'(Op, Type0, Type, Style) :-
	'$write_atom'(Op, Type0, Type, Style).

/*===============================================================
	Characcter Input/Output
===============================================================*/
nl :- nl.

nl(Stream) :- nl(Stream).
%nl(Stream) :- '$nl'(Stream), !.
%nl(Stream) :- '$illarg'(type(stream), nl(Stream), 1).

ttynl :- nl(user).

get0(X) :- get0(X).

get0(Stream, X) :- get0(Stream, X).

%get0(Stream, X) :- '$get0'(Stream, X1), !, X = X1.
%get0(Stream, X) :- '$illarg'(type(stream), get0(Stream, X), 1).

ttyget0(X) :- get0(user, X).

peek_char(X) :- '$get0'(X1), '$unget'(X1), X = X1.

peek_char(Stream, X) :- '$get0'(Stream, X1), !,
	'$unget'(Stream, X1), X = X1.
peek_char(Stream, X) :- '$illarg'(type(stream), peek_char(Stream,X), 1).

get(X) :- get(X).

get(Stream, X) :- get(Stream, X).
%get(Stream, X) :- '$get'(Stream, X1), !, X = X1.
%get(Stream, X) :- '$illarg'(type(stream), get(Stream, X), 1).

ttyget(X) :- get(user, X).

skip(X) :-
	I is X,
	(   I /\ -256 =\= 0 ->
	    '$illarg'(domain(integer,character), skip(X), 1)
	;   '$skip_until'(I)
	).

skip(Stream, X) :-
	I is X,
	(   I /\ -256 =\= 0 ->
	    '$illarg'(domain(integer,character), skip(Stream, X), 2)
	;   '$skip_until'(Stream, I) -> true
	;   '$illarg'(type(stream), skip(S,X), 1)
	).

ttyskip(X) :- skip(user, X).

'$skip_until'(X) :- 
	'$get0'(C), 
	(   X =:= C -> true
	;   '$skip_until'(X)
	).

'$skip_until'(Stream, X) :- 
	'$get0'(Stream, C),
	(   X =:= C -> true
	;   '$skip_until'(Stream, X)
	).

put(X) :- put(X).
%put(X) :- I is X, !, '$put'(I).

put(Stream, X) :- put(Stream, X).
%put(Stream, X) :- I is X, '$put'(Stream, I), !.
%put(Stream, X) :- '$illarg'(type(stream), put(Stream, X), 1).

ttyput(X) :- put(user, X).
%ttyput(X) :- I is X, '$put'(user, I).

tab(X) :- tab(X).
%tab(X) :- I is X, !, '$tab'(I).

tab(Stream, X) :- tab(Stream, X).
%tab(Stream, X) :- I is X, '$tab'(Stream, I), !.
%tab(Stream, X) :- '$illarg'(type(stream), tab(Stream,X), 1).

ttytab(X) :- I is X, !, '$tab'(user, I).

%tab(N) :- N =< 0, !.
%tab(N) :- put(32), N1 is N-1, tab(N1).
%tab(Stream, N) :- N =< 0, !.
%tab(Stream, N) :- put(Stream, 32), N1 is N-1, tab(Stream, N1).

/*===============================================================
	Stream IO
===============================================================*/
open(File, Mode, Stream) :- 
	nonvar(Mode), 
	'$open_mode'(Mode),
	!,
	'$fopen'(File, Mode, Stream).

open(File, Mode, Stream) :-
	findall(M, '$open_mode'(M), Domain),
	'$illarg'(domain(atom,Domain), open(File,Mode,Stream), 2).

'$fopen'(File, Mode, Stream) :- 
	'$open'(File, Mode, Stream0),
	!, 
	Stream = Stream0.
'$fopen'(File, Mode, Stream) :-
	'$illarg'(permission(open,file,File,0), open(File,Mode,Stream), 0).

'$open_mode'(read).
'$open_mode'(write).
'$open_mode'(append).

close(X) :- close(X).
%close(X) :- nonvar(X), '$close'(X), !.
%close(X) :- '$illarg'(type('file or stream'), close(X), 1).

current_input(X) :- current_input(X).

current_output(X) :- current_output(X).

set_input(X) :- set_input(X).
%set_input(X) :- '$set_input'(X), !.
%set_input(X) :-	'$illarg'(type(stream), set_input(X), 1).

set_output(X) :- set_output(X).
%set_output(X) :- '$set_output'(X), !.
%set_output(X) :- '$illarg'(type(stream), set_output(X), 1).

flush_output :- flush_output.

flush_output(X) :- flush_output(X).
%flush_output(X) :- '$flush_output'(X), !.
%flush_output(X) :- '$illarg'(type(stream), flush_output(X), 1).

ttyflush :- flush_output(user).

%'$builtin_stream'(user, read, Stream) :- !, 
%	'$standard_input'(Stream).
%'$builtin_stream'(user, write, Stream) :- !, 
%	'$standard_output'(Stream).
%'$builtin_stream'(user_input, read, Stream) :- !, 
%	'$standard_input'(Stream).
%'$builtin_stream'(user_output, write, Stream) :- !, 
%	'$standard_output'(Stream).
%'$builtin_stream'(user_error, write, Stream) :- !, 
%	'$standard_error'(Stream).

/*===============================================================
	Operators
===============================================================*/
op(Prec, Type, Op) :- integer(Prec), 0 =<Prec, Prec =<1200, !,
	'$op1'(Prec, Type, Op).
op(Prec, Type, Op) :- 
	'$illarg'(domain(integer,0-1200), op(Prec,Type,Op), 1).

'$op1'(Prec, Type, Op) :- nonvar(Type), '$op_type'(Type), !,
	'$op2'(Prec, Type, Op).
'$op1'(Prec, Type, Op) :- 
	findall(X, '$op_type'(X), Domain),
	'$illarg'(domain(term,Domain), op(Prec,Type,Op), 2).

'$op2'(Prec, Type, Op) :- '$op_atoms'(Op, Atoms, []), !, 
	'$op3'(Atoms, Prec, Type).
'$op2'(Prec, Type, Op) :- 
	'$illarg'(type(list(atom)), op(Prec,Type,Op), 3).

'$op3'(Atoms, 0, Type) :- !, '$retract_opts'(Atoms, Type).
'$op3'(Atoms, Prec, Type) :- '$assert_opts'(Atoms, Prec, Type).

'$retract_opts'([], _) :- !.
'$retract_opts'([A|As], Type) :-
	retractall('$current_op0'(_, Type, A)),
	'$retract_opts'(As, Type).

'$assert_opts'([], _, _) :- !.
'$assert_opts'([A|As], Prec, Type) :- 
	assert('$current_op0'(Prec, Type, A)),
	'$assert_opts'(As, Prec, Type).

'$op_atoms'([]) --> !.
'$op_atoms'([X|Xs]) --> !, {atom(X)}, [X], '$op_atoms'(Xs).
'$op_atoms'(X) --> '$op_atoms'([X]).

'$op_type'(fy).
'$op_type'(fx).
'$op_type'(yfx).
'$op_type'(xfy).
'$op_type'(xfx).
'$op_type'(yf).
'$op_type'(xf).

current_op(Prec, Type, Op) :-
	'$current_op0'(Prec, Type, Op).

% LLP operators
'$current_op0'( 1200, fx, (forall)).
'$current_op0'( 1190, xfx, (\)).
'$current_op0'( 1060, xfy, (&)).
'$current_op0'(  950, xfy, (-<>)).
'$current_op0'(  950, xfy, (=>)).
'$current_op0'(  900,  fy, (!)).
% SICStus Prolog operators
'$current_op0'( 1170, xfx, (:-)).
'$current_op0'( 1170, xfx, (-->)).
'$current_op0'( 1170,  fx, (:-)).
'$current_op0'( 1170,  fx, (?-)).
%'$current_op0'( 1200, xfx, (:-)).
%'$current_op0'( 1200, xfx, (-->)).
%'$current_op0'( 1200,  fx, (:-)).
%'$current_op0'( 1200,  fx, (?-)).
'$current_op0'( 1150,  fx, (mode)).
'$current_op0'( 1150,  fx, (public)).
'$current_op0'( 1150,  fx, (dynamic)).
'$current_op0'( 1150,  fx, (resource)).
'$current_op0'( 1150,  fx, (multifile)).
'$current_op0'( 1150,  fx, (block)).
'$current_op0'( 1150,  fx, (meta_predicate)).
'$current_op0'( 1100, xfy, (;)).
'$current_op0'( 1050, xfy, (->)).
'$current_op0'( 1000, xfy, (',')).
'$current_op0'(  900,  fy, (\+)).
'$current_op0'(  900,  fy, (spy)).
'$current_op0'(  900,  fy, (nospy)).
'$current_op0'(  700, xfx, (=)).
'$current_op0'(  700, xfx, (is)).
'$current_op0'(  700, xfx, (=..)).
'$current_op0'(  700, xfx, (==)).
'$current_op0'(  700, xfx, (\==)).
'$current_op0'(  700, xfx, (@<)).
'$current_op0'(  700, xfx, (@>)).
'$current_op0'(  700, xfx, (@=<)).
'$current_op0'(  700, xfx, (@>=)).
'$current_op0'(  700, xfx, (=:=)).
'$current_op0'(  700, xfx, (=\=)).
'$current_op0'(  700, xfx, (<)).
'$current_op0'(  700, xfx, (>)).
'$current_op0'(  700, xfx, (=<)).
'$current_op0'(  700, xfx, (>=)).
'$current_op0'(  550, xfy, (:)).
'$current_op0'(  500, yfx, (+)).
'$current_op0'(  500, yfx, (-)).
'$current_op0'(  500, yfx, (#)).
'$current_op0'(  500, yfx, (/\)).
'$current_op0'(  500, yfx, (\/)).
'$current_op0'(  500,  fx, (+)).
'$current_op0'(  500,  fx, (-)).
'$current_op0'(  400, yfx, (*)).
'$current_op0'(  400, yfx, (/)).
'$current_op0'(  400, yfx, (//)).
'$current_op0'(  400, yfx, (<<)).
'$current_op0'(  400, yfx, (>>)).
'$current_op0'(  300, xfx, (mod)).
'$current_op0'(  300, xfx, (~)).
'$current_op0'(  200, xfy, (^)).

/*===============================================================
	Utilities
===============================================================*/
'$member'(X, [X|_]).
'$member'(X, [_|L]) :- '$member'(X, L).


