/*
 * @(#)PrologInterface.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;
	using System.Reflection;

	public sealed class PrologInterface {
		private bool result;
		Predicate code = null;
		public Prolog engine;

		bool daemon = false;
		bool signalExit = false;

		public PrologInterface( ) {
			engine = new Prolog( );
		}

		public PrologInterface( bool signalExit ) {
			engine = new Prolog( new PushbackReader( Console.In ),
				Console.Out, Console.Error );
			daemon = true;
			this.signalExit = signalExit;
		}

		public PrologInterface( PushbackReader new_in, System.IO.TextWriter new_out,
			System.IO.TextWriter new_err, bool signalExit ) {
			engine = new Prolog( new_in, new_out, new_err );
			daemon = true;
			this.signalExit = signalExit;
		}

		public void setPredicate(Predicate p) {
			code = p;
		}

		public bool call() {
			lock( this ) {
				this.start();	 /* execute run() in new thread */
				try {
					Monitor.Wait(this);	     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}

		public bool redo() {
			lock( this ) {
				Monitor.Pulse(this);	    /* resume child thread */
				try {
					Monitor.Wait(this);     /* wait caller's thread */
				} catch (ThreadInterruptedException) {}
				return result;
			}
		}

		public void success() {
			lock( this ) {					   
				result = true;
				Monitor.Pulse(this);   	/* resume caller's thread */
				try {
					Monitor.Wait(this); 	/* wait child thread */
				} catch (ThreadInterruptedException) {}
			}
		}

		public void fail() {
			lock( this ) {
				result = false;
				Monitor.Pulse(this);	    /* resume caller's thread */
			}
		}

		private Thread t = null;

		public void start() {
			t = new Thread( new ThreadStart( run ) );
			if( daemon )
				t.IsBackground = true;
			t.Start( );
		}

		public void stop() {
			t.Abort( );
		}

		public void addCallingAssembly( ) {
			if( !Assembly.GetCallingAssembly( ).Equals( Assembly.GetExecutingAssembly( ) ) )
				AssemblyList.Add( Assembly.GetCallingAssembly( ), null );
		}

		public void run() {
			try {
				try {
					while (true) {
						if (code == null) break;
					
						// System.Console.WriteLine( code.ToString( ) );
						//code.setEngine(engine); // THIS CALL MADE A PARAMETER (Threading)
						code = code.exec( engine ); // engine );
						code = engine.ServiceQueue( code );
					}
				}
				catch (AbortException e){
					System.Console.WriteLine(e.Message);
				}
				fail();
				/* START JJC */
				if( signalExit )
					foreach( ConcurrentVariable cv in engine.concurrentVars )
						cv.ReportThreadDeath( engine );
				/* END JJC */
			} finally {
				engine.releaseAllLocks( );
			}
		}
	}
}