/*
 * @(#)VariableTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 */

// import java.util.*;

namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;

	public sealed class VariableTerm : Term {

		private static long ctr = Int64.MinValue;
				
		private static long incCounter( ) {
			lock( typeof( VariableTerm ) ) {
				return ctr++;
			}
		}

		/* START JJC - Added for fork/wait support */

		//internal volatile VariableTerm thread_root = null;
		//internal volatile VariableTerm engine_root = null;
		//internal System.Collections.ArrayList thread_children = 
		//	new System.Collections.ArrayList( );
		//internal CsObjectTerm thread_root_object = null;
		internal ConcurrentVariable concurrentRoot = null;
		Prolog engine = null;

		// added to support lock/unlock.

		internal object locked_object = null;
		internal object unlocked_object = null;

		/* END JJC */

		Term val;

		private long itimeStamp;
		public long timeStamp {
			get {
				return itimeStamp;
			}
			set {
				itimeStamp = value;
			}
		}
		private long ID;

		public VariableTerm(long t, Prolog engine) {
			//	val = null;
			this.engine = engine;
			val = this;
			timeStamp = t;
			ID = incCounter( );
		}

		public VariableTerm() {
			//	val = null;
			val = this;
			timeStamp = Int64.MinValue;
			ID = incCounter( ); 
		}

		public override Term dereference() {
			if (val == this)
				return this;
			return val.dereference();
		}

		//    public void undo() { val = null; }
		public void undo() { 
			val = this;
			if( locked_object != null ) {
				System.Threading.Monitor.Exit( locked_object );
				engine.reportUnlock( locked_object );
				//engine.user_output.WriteLine( "Unlocked (by undo) " + locked_object );
				locked_object = null;
			}
			if( unlocked_object != null ) {
				System.Threading.Monitor.Enter( unlocked_object );
				engine.reportLock( unlocked_object );
				//engine.user_output.WriteLine( "Locked (by undo) " + unlocked_object );
				unlocked_object = null;
			}
		}

		public override Term copyTerm(Prolog engine) {
			System.Collections.Hashtable table = engine.varTable;
			VariableTerm copy;
			if (val != this)
				return val.copyTerm(engine);
			copy = (VariableTerm)table[this];
			if ( copy == null ){
				copy = engine.makeVariable();
				table[ this ] = copy;
			}
			return copy;
		}

		/* START JJC */

		public override Term cloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			System.Collections.Hashtable cloneTable = oldEngine.cloneVarTable;
			//VariableTerm copy;
			if (val != this)
				return val.cloneTerm( oldEngine, newEngine, makeConc );
			//copy = (VariableTerm)table[this];

			VariableTerm clone = (VariableTerm)cloneTable[ this ]; // newEngine.makeVariable( );

			if ( clone == null ){ 
				clone = newEngine.makeVariable();
				cloneTable[ this ] = clone;
			}

			// copy the root to the new clone.
			if( makeConc ) {
				if( concurrentRoot == null ) {
					concurrentRoot = new ConcurrentVariable( );
					concurrentRoot.ReportFork( oldEngine );
					oldEngine.concurrentVars.Add( concurrentRoot );
					//thread_root_object = new CsObjectTerm( new RootSynchObj( ) );
				}
				clone.concurrentRoot = this.concurrentRoot;
				concurrentRoot.ReportFork( newEngine );
				newEngine.concurrentVars.Add( concurrentRoot );

				//this.thread_children.Add( clone );
			}

			/*
			if( copy.val == copy ) 
				clone.val = clone;
			else
				clone.val = copy.val.cloneTerm( oldEngine, newEngine, makeConc );
			*/	
			
			return clone;
		}
		/* END JJC */

		public bool OldUnify( Term t, Trail trail ) {
			VariableTerm v1;
			if (val != this)
				return val.unify(t, trail);
			
			t = t.dereference();
			if ( t.isVariable() ){
				VariableTerm v2 = (VariableTerm)t;
				if (this.timeStamp < v2.timeStamp ){
					v2.val = this;
					v1 = v2;
				} else {
					this.val = t;
					v1 = this;
				}
			} else {
				this.val = t; 
				v1 = this;
			}
			if ( v1.timeStamp < trail.engine.stack.getTimeStamp() ){
				trail.push(v1);
			}
			return true;
		}

		public override bool unify(Term t, Trail trail) {
			VariableTerm v1;
			if (val != this)
				return val.unify(t, trail);
			
			t = t.dereference();
			if ( t.isVariable() ){
				VariableTerm v2 = (VariableTerm)t;
				if (this.timeStamp < v2.timeStamp ){
					v2.val = this;

					/* START JJC */
					if( this.concurrentRoot != null )
						v2.concurrentRoot = this.concurrentRoot;
					/* END JJC */

					v1 = v2;
				} else {
					this.val = t; 
					
					/* START JJC */					
					if( v2.concurrentRoot != null )
						this.concurrentRoot = v2.concurrentRoot;
					/* END JJC */

					v1 = this;
				}
			} else {
				this.val = t;
				if( this.concurrentRoot != null ) {
					if( trail.engine.concInstQueue == null )
						trail.engine.concInstQueue = new System.Collections.Queue( );
					trail.engine.concInstQueue.Enqueue( 
						new NewBinding( this.concurrentRoot.csObject, t ) );
				}
				v1 = this;
			}
			if ( v1.timeStamp < trail.engine.stack.getTimeStamp() ){
				trail.push(v1);
			}
			return true;
		}

		public override bool Equals(Term t) {
			if( val != this )
				return val.Equals(t);
			//	t = t.dereference();
			if ( !t.isVariable() )
				return false;
			return this == t;
			//	return this.timeStamp == ((VariableTerm)t).timeStamp();
		}

		/*
		public Class getJavaType() { 
		if (val != null)
			return val.getJavaType();
		return java.lang.Object.class;
		}*/
		public override object toCsObject() {
			if (val != this)
				return val.toCsObject();
			return this;
			//	return new CsObjectTerm(this);
		}

		public String name() {
			//return "_" + timeStamp();
			return "_" + this.GetHashCode( ).ToString( );
		}

		public long getID() {
			return ID;
		}

		public override String ToString() {
			if (val != this)
				return val.ToString();
			return name();
		}

		/* START JJC (for interop from C# side) */

        //private static PrologInterface pi = new PrologInterface( );

		public void Send( Term inst ) {
			PrologInterface pi = new PrologInterface( false );
			this.undo( );
			pi.setPredicate( new Predicates.dollar_unify_2( this, inst, null ) );
			pi.run( );
		}

		public Term Receive( ) {
			PrologInterface pi = new PrologInterface( false );
			this.undo( );
			pi.setPredicate( new Predicates.WaitForOne_1( this, null ) );

			//pi.setPredicate( new Predicates.dollar_forkBlock_1( 
			//	new StructureTerm( 
			//		SymbolTerm.makeSymbol( "wait_for", 1 ),
			//	    new Term[] { this } ), null ) );

			pi.run( );
			return this;
		}

		/* END JJC */

	}
}


